require("log-timestamp");
const { System } = require("syrus4-nodejs");
const storage = require("./services/storage");
const { structure, destinations, syrusjs_variables, addCallbackToSignals } = require("./scripts/globals");
const { getConfiguration, getCommandsFromconfigurationFiles } = require("./scripts/configuration");
const { createDestination } = require("./scripts/destinations");
const { processCommand, fireSignal } = require("./scripts/processor");

var configuration = { namespace: "default" };

/**
 * Event Machine
 */

 /**
 * Signals evaluator for check if the signal is up or down (True|False)
 * @param {Array<signal>} signals all the signals to be evaluated when some data changed
 */
function evaluateSignals(signals, path) {
	signals.forEach(signal => {
		try {
			let { variable, comparator, operator, name, min_duration } = signal;
			if (variable.includes("*")) variable = path
			if (!min_duration) min_duration = 0;
			signal.__previous_state = signal.state;
			signal.__transition_state = _evaluator(variable, operator, comparator);
			// if the signal is different from the previous indicates that the signal must fire
			if (signal.__transition_state != signal.__previous_state) {
				clearTimeout(signal.__timeout);
				// Timeout to wait when signal is declared within
				signal.__timeout = setTimeout(() => {
					if (signal.state != signal.__transition_state) {
						signal.state = signal.__transition_state;
						storage.hset("signals_states", name, signal.state);
						fireSignal(name);
					}
				}, min_duration);
				// Even is min_duration is 0 this help to deferred evaluation at the end of event loop
				// allowing evaluate all the signals first before fire the actions/events
			}
		} catch (error) {
			console.error(error);
			return;
		}
	});
}

/**
 *  Simple evaluator to evaluate signals  as math oprations
 * @param {*} variable variable on  the left
 * @param {*} op operator (<, <= , > , >=, != , ==)
 * @param {*} comparator  variable on the right
 */
function _evaluator(variable, op, comparator) {
	let var1 = syrusjs_variables[`${variable}`];
	// if there is no operator, use the variable value, if boolean, return that value, otherwise allow for changes
	if (!op && !comparator) {
		return !!var1;
	}
	var var2 = comparator.startsWith("$") || comparator.startsWith("@") ?
		syrusjs_variables[comparator] : comparator == "true" ?
		true : comparator == "false" ?
		false : !isNaN(Number(comparator)) ?
			Number(comparator) : comparator == "null" ?
			null : comparator;

	if (op == "==" || op == "===") {
		return var1 == var2;
	}
	if (op == ">") {
		return var1 > var2;
	}
	if (op == ">=") {
		return var1 >= var2;
	}
	if (op == "<") {
		return var1 < var2;
	}
	if (op == "<=") {
		return var1 <= var2;
	}
	if (op == "!=" || op == "<>" || op == "!==") {
		return var1 != var2;
	}
	throw "bad signal definition: " + variable + op + comparator;
}

// Added callback for when internal variable change from syrusjs
addCallbackToSignals("evaluate_signals", path => {
	const signals = Object.values(structure.signals).filter((sig) => {
		if (sig.variable.includes('*')){
			const [start, end] = sig.variable.split('*')
			return path.startsWith(start) && path.endsWith(end)
		}
		return sig.variable == path || sig.comparator == path
	});
	evaluateSignals(signals, path);
});

/**
 * the entry point of the app
 *
 */
async function init() {
	// Get all the basic information about the S4 and modem
	try {
		var [ system, modem ] = await Promise.all([ System.info(), System.modem() ]);
		syrusjs_variables["$system.version"] = global.OS_VERSION = system.apexVersion.replace("apex-", "");
		for (const key in modem) {
			syrusjs_variables[`$modem.${key.toLowerCase()}`] = modem[key];
		}
	} catch (error) {
		syrusjs_variables["$system.version"] = "20.0.0";
	}

	// Get the configuration from syrus.conf file, also the .configuration.json from the app_data folder and the json Schema
	configuration = await getConfiguration();
	// from configuration get all the syruslang commands to be proccesed
	var commands = getCommandsFromconfigurationFiles(configuration);
	console.log("active configuration:", configuration);
	console.log("listening for geofences namespaces:", configuration.namespace);

	console.log("processing commands...")
	for (const command of commands) {
		// console.log(order)
		let resp = await processCommand(command).catch(console.error);
		if(resp) console.warn( {cmd:`"${command.raw}"`, error: resp})
	}
	console.log("processing commands... DONE")

	/**
	 * here are stored the last state of every signal when app is restarted must preserve last state to prevent fire event twice
	*/
	console.log("loading signals...")
	var signals_states = await storage.hgetall("signals_states");
	for (const key in structure.signals) {
		const signal = structure.signals[key];
		if (signals_states[key] !== undefined) signal.state = signals_states[key] == "true";
	}
	console.log("loading signals... DONE")

	// set the watchers
	console.log("initializing watchers...")
	require("./scripts/watchers")(configuration);
	console.log("initializing watchers... DONE")

	// Once all the syruslang is proccesed the destination must be created.
	// this is no made in syruslang process time because sometimes appends.syrus.conf redefine some of the destination
	// and could lead to memory leaks and weird states on the destinations
	console.log("creating destinations...")
	for (const name in structure.destinations) {
		var config = structure.destinations[name];
		destinations[name] = createDestination(config, async params => {
			params.destination = config.name;
			var resp = await processCommand(params, true);
			if (!resp) resp = "OK";
			return resp;
		});
	}
	console.log("creating destinations... DONE")

	// socket_connection.js create a unix socket to comunicate with the app using syruslang
	// also part of the missing feature `syrus-apps-manager send` to comunicate between apps inside of the Syrus4
	console.log("initializing interprocess socket...")
	await require("./services/socket_connection")(processCommand).catch(console.error);
	console.log("initializing interprocess socket... DONE")
}

require("./scripts/errors-handler");

// start the application
init()
.then(()=>{
	console.log("running...")
})
.catch((error)=>{
	console.error(error)
	process.exit(1)
});

// TODO: Check syruslang config has a complete valid configuration
// TODO: improve buffer handler
// TODO: apx-mdm send_sms
