const fs = require("fs");
const { IOS, Geofences, Utils } = require("syrus4-nodejs");
const postFixEvaluator = require("../services/PostFix");
const { structure, syruslangParser, app_data_folder, destinations, syrusjs_variables, raters } = require("./globals");
const { createTrackingResolution, changeTrackingResolution, deleteTrackingResolution } = require("./tracking_resolutions");
const { createCounters, startCounters, stopCounters, deleteCounters } = require("./counters");
const { createTimer, startTimer, cancelTimer, deleteTimer } = require("./timers");
const { createDestination } = require("./destinations");
var tracking_resolution = {};

/**
 * Designed to fire internal signals like @tracking_resolution.signal or @power_save.sleep
 * creates the signal as no previously defined
 *
 * @param {*} path the name of the signal
 */
function fireInternalSignal(path) {
	if (!structure.signals[path]) {
		structure.signals[path] = {
			name: path,
			variable: path,
			operator: null,
			comparator: null,
			min_duration: 0,
			limit: 0,
			active: true
		};
	}
	syrusjs_variables[path] = true;
	setTimeout(() => {
		syrusjs_variables[path] = false;
	}, 20); // Arbitrary time to keep the signal on, this allow internal signal combination when trigger depends from  more than one internal signal,  both are to be actived at the same time.
}

/**
 * called when am action must execute his action
 * also evaluated the rater to prevent to be fired too much
 * @param {*} action the action to be triggered
 * @return {*}
 */
function triggerAction(action) {
	if (action.rate) {
		if (!raters[`action_${action.name}`]) {
			raters[`action_${action.name}`] = 0;
			setTimeout(() => {
				delete raters[`action_${action.name}`];
			}, action.rate.timeout);
		}

		// Check rater of the action (ex: rater=5/1min)
		raters[`action_${action.name}`]++;
		if (raters[`action_${action.name}`] > action.rate.times) {
			return console.error(`rate limit reached action ${action.name} ${raters[`action_${action.name}`]}/${action.rate.times}`);
		}
	}
	console.log("trigger action", action.name);
	var commands = syruslangParser.process(action.action);
	for (const command of commands) {
		console.log(command)
		// normally action contains another syruslang command
		processCommand(command).catch((error)=>{
			console.error("ERROR: executing action", command)
			console.error(error)
		});
	}
}

/**
 * fired everytime the signal goes from false -> true or vicerversa
 * here are fired all the associated actions and events to the signal
 * @param {*} signal_event
 */
function fireSignal(signal_event) {
	// console.log("signal changed:", signal_event);
	var values = {};
	Object.values(structure.signals).forEach(signal => {
		values[signal.name] = signal.state;
	});

	// Events to be sent
	var events = Object.values(structure.events).filter(event => {
		return event.signals.indexOf(signal_event) > -1;
	});
	events.forEach(event => {
		var result = postFixEvaluator(event.signals, values); // Make sure the trigger postfix is on true
		if (result) {
			console.log("fire event", event.name);
			sendEvent(event);
		}
	});

	// Actions to be triggered
	var actions = Object.values(structure.actions).filter(action => {
		return action.signals.indexOf(signal_event) > -1;
	});
	actions.forEach(action => {
		var result = postFixEvaluator(action.signals, values); // MAke sure the trigger postfix is on true
		if (result) triggerAction(action);
	});
}

/**
 * called when an event must be sent to the destinations
 * also evaluated the rater to prevent to be fired too much
 * @param {*} event the event to be triggered
 * @return {*}
 */
function sendEvent(event) {
	if (event.rate) {
		if (!raters[`event_${event.name}`]) {
			raters[`event_${event.name}`] = 0;
			setTimeout(() => {
				delete raters[`event_${event.name}`];
			}, event.rate.timeout);
		}
		raters[`event_${event.name}`]++;
		if (raters[`event_${event.name}`] > event.rate.times) {
			return console.error(`rate limit reached ${event.name} ${raters[`event_${event.name}`]}/${event.rate.times}`);
		}
	}
	// Event has fieldset, destinations and groups
	let dests = structure.groups[event.group].destinations;
	let fields = structure.fieldsets[event.fieldset].fields;
	let group = structure.groups[event.group];

	// If the event is own defined destinations, are used
	// else the group must define the destinations
	if (event.destinations && event.destinations.length > 0) dests = event.destinations;
	else if (group && group.destinations && group.destinations.length > 0) dests = group.destinations;

	// If the event is own defined field, this is  used
	// else the group must define the fieldset
	if (event.fieldset && structure.fieldsets[event.fieldset] && structure.fieldsets[event.fieldset].fields) fields = structure.fieldsets[event.fieldset].fields;
	else if (group.fieldset && structure.fieldsets[group.fieldset] && structure.fieldsets[group.fieldset].fields) fields = structure.fieldsets[group.fieldset].fields;

	if (!Array.isArray(dests)) dests = [ dests ];
	var variables = JSON.parse(JSON.stringify(syrusjs_variables));

	for (const destination of dests) {
		let dest = destinations[destination];
		if (dest) dest.sendEvent(variables, event, fields, event.label, event.code);
		else console.error("undefined destination point", destination);
	}
}

/**
 * replaces from an text (normally syruslang) variables from syrusjs
 *	example: define destination mqtt json mqtt://mqtt.example.io:1883 protocol="mqtt" subscribe="dev/{{$modem.imei}|/messagesrx" publish="dev/{{$modem.imei}}/messagestx"
 * @param {*} txt syruslang command that has the curly brackects
 * @return {*} the txt with the variable already replaced
 */
function replaceArguments(txt) {
	while (txt.indexOf("{{") > -1 && txt.indexOf("}}") > -1) {
		var start = txt.indexOf("{{");
		var end = txt.indexOf("}}");
		var variable = txt.substring(start + 2, end);
		txt = txt.replace(`{{${variable}}}`, syrusjs_variables[variable]);
	}
	return txt;
}

/**
 * this is the main evaluator of command, syruslang is converted to json actions object which define how syrusjs must act.
 *
 * @param {*} data the json action object, the .command indicates the order, very similiar as REDUX works in reactjs, or at least that's the idea
 * @param {boolean} [shouldAppend=false] if this on true and the command is a definition/set the command is added to the appends file, this only should be true when the function is called from a remote destination.
 * @return {*} is the response contains a valid text as response like get signal ignON ==> define signal ignOn $io.ign == true
 */
async function processCommand(data, shouldAppend = false) {
	try {
		for (const key in data) {
			if (![ "command", "action" ].includes(key) && typeof data[key] == "string") {
				data[key] = replaceArguments(data[key]);
			}
		}
		if (data.isSyrusLang) data = syruslangParser.process(data.command)[0];

		if (data.command == "exec") {
			var command = replaceArguments(data.argument).trim();
			if (command.startsWith("speak") || command.startsWith("apx-tts")) {
				command = command.replace(/_/g, " ");
			}
			return await Utils.OSExecute(`${command}`);
		} else if (data.command == "define destination") {
			structure.destinations[data.name] = data;
			// Avoid creating destinations while syruslang config still procesing
			// sometimes destinations are redefined in the appends and could leaks to errors
			if (shouldAppend)
				destinations[data.name] = createDestination(data, async params => {
					params.destination = data.name;
					var resp = await processCommand(params, true);
					if (!resp) resp = "OK";
					return resp;
				});
		} else if (data.command == "define tracking_resolution") {
			structure.tracking_resolutions[data.name] = data;
			createTrackingResolution(data);
		} else if (data.command == "define signal") {
			data.active = true;
			structure.signals[data.name] = data;
		} else if (data.command == "define event") {
			structure.events[data.name] = data;
		} else if (data.command == "define group") {
			structure.groups[data.name] = data;
		} else if (data.command == "define fieldset") {
			structure.fieldsets[data.name] = data;
		} else if (data.command == "define action") {
			structure.actions[data.name] = data;
		} else if (data.command == "define geofence") {
			// structure.geofences[data.name] = data;
			// structure.geofences[data.name].raw = data.raw;
			if (data.group.length > 0) {
				for (const group of data.group) {
					if (!structure.collections) {
						structure.collections = {};
					}
					if (!structure.collections[group]) {
						structure.collections[group] = [];
					}
					structure.collections[group].push(data.name);
				}
			}
			await Geofences.addGeofence(data);
		} else if (data.command == "define counters") {
			structure.counters[data.name] = data;
			createCounters(data);
		} else if (data.command == "define timer") {
			createTimer(data)
		} else if (data.command == "start counters") {
			startCounters(data.name);
		} else if (data.command == "stop counters") {
			stopCounters(data.name);
		} else if (data.command == "toggle") {
			structure[`${data.set}s`][data.name].active = data.state;
		} else if (data.command == "set variable") {
			structure.variables[data.name] = data;
			syrusjs_variables[`$variables.${data.name}`] = data.value;
		} else if (data.command == "set destinations") {
			var key = data.group ? "group" : data.event ? data.event : null;
			if (key) {
				if (!structure.groups[data[key]]) {
					return `undefined group ${data[key]}`;
				}
				structure.groups[data[key]].destinations = data.destinations;
			}
		} else if (data.command == "set namespace") {
			structure.namespace = data.name;
		} else if (data.command == "set out") {
			IOS.setOutputState(`OUT${data.io}`, data.state == "on");
		} else if (data.command == "set tracking_resolution") {
			tracking_resolution = structure.tracking_resolutions[data.name];
			if (!tracking_resolution) {
				return console.error("undefined tracking_resolution " + data.name);
			}
			changeTrackingResolution(data);
		} else if (data.command == "set value") {
			if (data.variable_name.startsWith("@")) {
				return "can't set that variable:" + data.variable_name;
			}
			syrusjs_variables[data.variable_name] = data.value;
		} else if (data.command == "get") {
			return await getDataSyrusLang(data);
		} else if (data.command == "enable" || data.command == "start") {
			if (data.type == "destination") {
				if (!destinations[data.name]) {
					return "undefined destination " + data.name;
				}
				destinations[data.name].client.enable();
			}
			if (data.type == "timer") {
				if (!structure.timers[data.name]) {
					return "undefined timer " + data.name;
				}
				startTimer(data.name);
			}
		} else if (data.command == "disable" || data.command == "stop") {
			if (data.type == "destination") {
				if (!destinations[data.name]) {
					return "undefined destination " + data.name;
				}
				destinations[data.name].client.disable();
			}
			if (data.type == "timer") {
				if (!structure.timers[data.name]) {
					return "undefined timer " + data.name;
				}
				cancelTimer(data.name);
			}
		} else if (data.command == "delete") {
			let DELETABLE = [
				"destination", "namespace", "tracking_resolution",
				"signal", "event", "group",
				"action", "fieldset", "geofence", "counters",
				"timer"
			]
			if (!DELETABLE.includes(data.type)) {
				return "unknown type " + data.type;
			}
			if (data.type == "counters") data.type = "counter";
			let cmd = structure[data.type + "s"][data.name];
			if (cmd) {
				if (data.type == "destination") {
					cmd.disconnect();
				}
				if (data.type == "tracking_resolution") {
					deleteTrackingResolution(data.name);
				}
				if (data.type == "counter") {
					deleteCounters(data.name);
				}
				if (data.type == "geofence") {
					Geofences.removeGeofence(data);
				}
				if (data.type == "timer") {
					deleteTimer(data)
				}
				delete structure[data.type + "s"][data.variable_name];
			}
		} else if (data.command == "send event") {
			let event = structure.events[data.name];
			if (!event && data.code) event = Object.values(structure.events).find(ev => ev.code == data.code);
			if (!event && data.label) event = Object.values(structure.events).find(ev => ev.label == data.label);
			if (!event) return `undefined event ${data.name}`;
			sendEvent(event);
		} else if (data.command == "trigger action") {
			let action = structure.actions[data.name];
			if (!action) return `undefined action ${data.name}`;
			triggerAction(action);
		} else if (data.command == "fire signal") {
			fireInternalSignal(data.signal_name);
		}
		if (shouldAppend && data.persistent) {
			let appendsPath = `${app_data_folder}/appends.syrus.conf`;
			fs.appendFileSync(appendsPath, "\n" + data.raw);
		}
	} catch (error) {
		console.error({cmd: data.raw, error: error});
	}
}

/**
 * function to get the definitions of the syruslang configuration
 *
 * @param {*} data the json action from processCommand
 * @return {*} the current definition(s), that is being asked for.
 */
async function getDataSyrusLang(data) {
	if ([ "signals", "events", "actions", "destinations", "tracking_resolutions", "fieldsets", "variables" ].includes(data.type))
		return Object.values(structure[data.type]).map(signal => signal.raw).join(";");
	if (
		![
			"value",
			"destination",
			"destination_state",
			"destinations",
			"namespace",
			"tracking_resolution",
			"variable",
			"signal",
			"event",
			"group",
			"action",
			"fieldset",
			"geofence",
			"counters"
		].includes(data.type)
	) {
		return "unknown type " + data.type;
	}

	if (data.type == "counters") data.type = "counter";
	if (data.type == "variable") return `set variable ${data.variable_name} ${syrusjs_variables[`$variables.${data.variable_name}`]}`;

	if (data.type == "destinations") {
		let cmd = structure.groups[data.variable_name];
		if (!cmd) {
			return `unknown group destination: ${data.variable_name}`;
		}
		return `set ${data.type} group=${data.variable_name} ${cmd.destinations.join(" ")}`;
	}

	if (data.type == "destination_state") {
		let dest = destinations[data.variable_name];
		if (!dest) {
			return `unknown destination: ${data.variable_name}`;
		}
		return `${dest.client.active ? "enable" : "disable"}`;
	}

	if (data.type == "value") {
		return syrusjs_variables[data.variable_name] !== undefined ? syrusjs_variables[data.variable_name] : "undefined";
	}

	if (data.type == "geofence") {
		try{
			return await Geofences.get({ name: data.variable_name });
		}catch(error){
			console.error(error)
			return `undefined geofence ${data.variable_name}`
		}
	}

	var cmd = structure[data.type + "s"][data.variable_name];
	if (!cmd) {
		return `unknown ${data.type}: ${data.variable_name}`;
	}
	if (data.returnValue) {
		return `${data.type} value=${cmd.state}`;
	} else {
		return cmd.raw;
	}
}

module.exports = {
	processCommand,
	triggerAction,
	sendEvent,
	fireSignal,
	replaceArguments
}
