const { GPS } = require("syrus4-nodejs");
const dotted = require("dot-object");
const { tracks, syrusjs_variables, structure } = require("./globals");
const dayjs = require("dayjs").extend(require("dayjs/plugin/utc")).extend(require("dayjs/plugin/localizedFormat"));
require("dayjs/locale/es");

/**
 * Called when a tracking resolution must be created
 * normally called when syruslang json action indicates "define tracking_resolution"
 * @param {*} resolution the resolution parameters
 */
async function createTrackingResolution(resolution) {
	return tracks[resolution.name] = GPS.watchTrackingResolution(
		(data, reason) => {
			onLocation(data, reason, resolution.name);
		},
		{
			distance: resolution.distance,
			time: resolution.time / 1000,
			heading: resolution.heading,
			namespace: resolution.name
		}
	);
}

/**
 * Called when a tracking resolution must be redefined
 * normally called when syruslang json action indicates "set tracking_resolution"
 * @param {*} resolution the new resolution parameters
 */
async function changeTrackingResolution(resolution) {
	if (await GPS.getTrackingResolution({ namespace: resolution.name })) {
		return GPS.setTrackingResolution({
			distance: resolution.distance,
			time: resolution.time / 1000,
			heading: resolution.heading,
			namespace: resolution.name
		});
	} else {
		return `undefined tracking resolution ${resolution.name}`;
	}
}

/**
 * Called when a tracking resolution must be deleted
 * normally called when syruslang json action indicates "delete tracking_resolution"
 * @name {*} name of the resolution to be deleted
 */
async function deleteTrackingResolution(name) {
	if (tracks[name]) await tracks[name].unsubscribe();
	delete tracks[name];
}

/**
 * Callback to execute when tracking resolution delivers new gps data
 * @param {*} gps gps where is all data as Geolocation coordinates
 * @param {*} reason the reason why the tracking_resolution was fired possible 'time','heading','distance'
 * @param {*} namespace name of the tracking_resolution
 * @param {*} shouldFireSignal if the location should fire the internal signals @tracking_resolution.namespace.reason, always true for now
 */
function onLocation(gps, reason, namespace, shouldFireSignal = true) {
	Object.assign(syrusjs_variables, dotted.dot({ $gnss: gps.coords }));
	syrusjs_variables["$gnss.heading"] = gps.coords.bearing;
	syrusjs_variables["$gnss.mph"] = gps.coords.speed * 2.23694;
	syrusjs_variables["$gnss.kph"] = gps.coords.speed * 3.6;
	syrusjs_variables["$gnss.timestamp"] = gps.timestamp;

	syrusjs_variables["$speed"] = Math.round(gps.coords.speed);
	syrusjs_variables["$mph"] = Math.round(gps.coords.speed * 2.23694);
	syrusjs_variables["$kph"] = Math.round(gps.coords.speed * 3.6);
	syrusjs_variables["$timestamp"] = dayjs.unix(gps.timestamp).locale("en").format("LLLL");
	syrusjs_variables["$tiempo"] = dayjs.unix(gps.timestamp).locale("es").format("LLLL");
	Object.assign(syrusjs_variables, dotted.dot({ $gnss: gps.extras }));
	if (shouldFireSignal) {
		fireInternalSignal(`@tracking_resolution.${namespace}.signal`);
		if (reason != "signal") fireInternalSignal(`@tracking_resolution.${namespace}.${reason}`);
	}
}

/**
 * Designed to fire internal signals like @tracking_resolution.signal or @power_save.sleep
 * creates the signal as no previously defined
 *
 * @param {*} path the name of the signal
 */
function fireInternalSignal(path) {
	if (!structure.signals[path]) {
		structure.signals[path] = {
			name: path,
			variable: path,
			operator: null,
			comparator: null,
			min_duration: 0,
			limit: 0,
			active: true
		};
	}
	syrusjs_variables[path] = true;
	setTimeout(() => {
		syrusjs_variables[path] = false;
	}, 20); // Arbitrary time to keep the signal on, this allow internal signal combination when trigger depends from  more than one internal signal,  both are to be actived at the same time.
}

module.exports = {
	createTrackingResolution,
	changeTrackingResolution,
	deleteTrackingResolution
};
