// handlers for the watchers, to check changes on S4, the
const dayjs = require("dayjs")
	.extend(require("dayjs/plugin/utc"))
	.extend(require("dayjs/plugin/localizedFormat"));
require("dayjs/locale/es");
const fs = require("fs");
const dotted = require("dot-object");

const { GPS, IOS, System, Network} = require("syrus4-nodejs");
const { Accelerometer, ECU, Geofences } = require("syrus4-nodejs");
const { IButton, Temperature } = require("syrus4-nodejs");
const { Serial } = require("syrus4-nodejs");

const { syrusjs_variables, structure } = require("./globals");

// sets all keys on an object as null
function clear_object(object){
	for (const key in object) {
		object[key] = null
	}
}

/**
 * Designed to fire internal signals like @tracking_resolution.signal or @power_save.sleep
 * creates the signal as no previously defined
 *
 * @param {*} path the name of the signal
 */
function fireInternalSignal(path) {
	if (!structure.signals[path]) {
		structure.signals[path] = {
			name: path,
			variable: path,
			operator: null,
			comparator: null,
			min_duration: 0,
			limit: 0,
			active: true
		};
	}
	syrusjs_variables[path] = true;
	setTimeout(() => {
		syrusjs_variables[path] = false;
	}, 20);
	// Arbitrary time to keep the signal on, this allowd internal signals to combine when the trigger depends on multiple internal signals.
}

// TODO: don't initialize all watchers unless required,
// this is a sideffect from favoring TAIP, that requires certain watchers always running
module.exports = function handler(configuration) {
	function onGpsData(gps) {
		Object.assign(syrusjs_variables, dotted.dot({ $gnss: gps.coords }));
		syrusjs_variables["$gnss.heading"] = gps.coords.bearing;
		syrusjs_variables["$gnss.mph"] = gps.coords.speed * 2.23694;
		syrusjs_variables["$gnss.kph"] = gps.coords.speed * 3.6;
		syrusjs_variables["$gnss.timestamp"] = gps.timestamp;

		syrusjs_variables["$speed"] = Math.round(gps.coords.speed);
		syrusjs_variables["$mph"] = Math.round(gps.coords.speed * 2.23694);
		syrusjs_variables["$kph"] = Math.round(gps.coords.speed * 3.6);
		// TODO: don't let the language go through to the variables(i.e $timestamp_es)
		syrusjs_variables["$timestamp"] = dayjs
			.unix(gps.timestamp)
			.locale("en")
			.format("LLLL");
		syrusjs_variables["$tiempo"] = dayjs
			.unix(gps.timestamp)
			.locale("es")
			.format("LLLL");
		Object.assign(syrusjs_variables, dotted.dot({ $gnss: gps.extras }));
	}
	function onGeofenceChange(fences) {
		if (!Array.isArray(fences)) {
			fences = [fences];
		}

		for (const fence of fences) {
			syrusjs_variables[`$geofences.${fence.name}.inside`] = fence.is_inside;
			syrusjs_variables[`$geofences.${fence.name}.time`] = fence.time || fence.timestamp;

			syrusjs_variables[`@last_geofence.name`] = fence.name;
			syrusjs_variables[`@last_geofence.state`] = fence.is_inside ? "inside" : "outside";
			syrusjs_variables[`@last_geofence.time`] = fence.time || fence.timestamp;
			fireInternalSignal(`@geofence.${fence.is_inside ? "enter" : "exit"}`);
		}
	}
	function onGeofenceGroupChange(groups) {
		if (!Array.isArray(groups)) {
			groups = [groups];
		}

		for (const group of groups) {
			syrusjs_variables[`$geofences.$groups.${group.name}.inside`] = group.is_inside;
			syrusjs_variables[`$geofences.$groups.${group.name}.time`] = group.time || group.timestamp;

			syrusjs_variables[`@last_geofence.group.name`] = group.name;
			syrusjs_variables[`@last_geofence.group.state`] = group.is_inside ? "inside" : "outside";
			syrusjs_variables[`@last_geofence.group.time`] = group.time;
			fireInternalSignal(`@geofence.group.${group.is_inside ? "enter" : "exit"}`);
		}
	}
	function onNetworkChange() {
		Network.getNetworks()
			.then(networks => {
				for (const net in networks) {
					const network = networks[net];
					var netname = net;
					switch (net) {
						case "wlan0":
							netname = "wifi";
							break;
						case "eth0":
							netname = "eth";
							break;
						case "ppp0":
							netname = "cell";
							break;
						default:
							break;
					}
					for (const key in network) {
						syrusjs_variables[`$net_${netname}.${key}`] = network[`${key}`];
					}
				}
			})
			.catch(console.error);
	}
	function onInputChange(ios) {
		for (const io in ios) {
			if (io == "BAT") {
				onBatteryChange(Number(ios[io]));
			} else {
				syrusjs_variables[`$io.${io.toLowerCase()}`] = ios[io];
			}
		}
	}
	function onBatteryChange(vdc) {
		/**
		 * battery level between 4000 and 3500
		 * keys: "$battery.level" , "$battery.voltaje"
		 * bat full 4000 >= 100%
		 * bat low  3620 <= 20%
		 * bat cri  3560 <= 10%
		 */
		if (vdc == null) {
			syrusjs_variables[`$battery.connected`] = false;
			syrusjs_variables[`$battery.voltage`] = 0;
			syrusjs_variables[`$battery.mv`] = 0;
			syrusjs_variables[`$battery.level`] = 0;
			return;
		}
		syrusjs_variables[`$battery.connected`] = true;
		syrusjs_variables[`$battery.voltage`] = vdc;
		syrusjs_variables[`$battery.mv`] = Math.round(vdc * 1000);
		var level = Math.round(((vdc * 1000 - 3500) / 500) * 100);
		if (level > 100) level = 100;
		if (level < 0) level = 0;
		if (syrusjs_variables[`$battery.level`] < 100 && level >= 100) {
			fireInternalSignal("@battery.full");
		}
		if (syrusjs_variables[`$battery.level`] > 20 && level <= 20) {
			fireInternalSignal("@battery.low");
		}
		if (syrusjs_variables[`$battery.level`] > 10 && level <= 10) {
			fireInternalSignal("@battery.critical");
		}
		syrusjs_variables[`$battery.level`] = level;
	}
	function onMotionChange(isMoving) {
		syrusjs_variables["$accelerometer.motion"] = isMoving;
	}
	function onMotionEvent(event, value, time = null) {
		syrusjs_variables[`$accelerometer.${event}.time`] = time ? time : new Date().toISOString();
		syrusjs_variables[`$accelerometer.${event}.value`] = value;
		fireInternalSignal(`@accelerometer.${event}.signal`);
	}
	const tires_dotted = {}
	function onEcuParamsChange(ecu_params) {
		let flat_data = {}
		for (const key in ecu_params) {
			let value = ecu_params[key]
			if (key.startsWith('$')){
				flat_data[key] = value
			}
			syrusjs_variables[`$ecu.${key}`] = value
		}
		let sub_data = dotted.object(flat_data)
		let tires = sub_data['$tires']
		if(tires){
			let all_tires = []
			let tire_signals = {
				"alarm_ok": [],
				"sensor_mute": [],
				"sensor_defective": [],
				// "sensor_loose": [],
				// "sensor_turned": [],
				// "sensor_battery_warning": [],
				// "sensor_over_temperature": [],
				"tire_leak": [],
				"temperature_warning": [],
				"over_inflation_critical": [],
				"over_inflation_warning": [],
				"under_inflation_warning": [],
				"under_inflation_critical": [],
			}
			for (const axle in tires){
				let axle_data = tires[axle]
				for (const tire in axle_data){
					let tire_data = axle_data[tire]
					let tire_ok = true
					tire_data.axle_index = axle.replace("axle_", "")
					tire_data.axle = parseInt(tire_data.axle_index) + 1
					tire_data.tire = tire.replace("tire_", "")
					all_tires.push(tire_data)
					switch (tire_data.sensor_status) {
						case 0:
							tire_ok = false
							tire_signals['sensor_mute'].push(tire_data)
							break
						case 2:
							tire_ok = false
							tire_signals['sensor_defective'].push(tire_data)
							break
					}
					if (tire_data.air_leak_status == 1) {
						tire_ok = false
						tire_signals['tire_leak'].push(tire_data)
					}
					if (tire_data.temperature_state == 1){
						tire_ok = false
						tire_signals['temperature_warning'].push(tire_data)
					}
					switch (tire_data.pressure_state){
						case 0:
							tire_ok = false
							tire_signals['over_inflation_critical'].push(tire_data)
							break
						case 1:
							tire_ok = false
							tire_signals['over_inflation_warning'].push(tire_data)
							break
						case 3:
							tire_ok = false
							tire_signals['under_inflation_warning'].push(tire_data)
							break
						case 4:
							tire_ok = false
							tire_signals['under_inflation_critical'].push(tire_data)
							break
					}
					if (tire_ok){
						tire_signals['alarm_ok'].push(tire_data)
					}
				}
			}
			let tires_update = {}
			for(const signal in tire_signals){
				let tires = tire_signals[signal]
				tires_update[`$tpms.${signal}`] = tires
				if (tires.length) fireInternalSignal(`@tpms.${signal}`)
			}
			// tires_data
			clear_object(tires_dotted)
			Object.assign(tires_dotted, dotted.dot(tires_update))
			console.log(tires_dotted)
			Object.assign(syrusjs_variables, tires_dotted);
		}
	}
	function onSleepOn() {
		fireInternalSignal(`@power_save.sleep`);
	}

	// TAIP ex
	// IB:$ibutton.connected.id,IS:$ibutton.last.id
	// IB:$ibutton.authorized.connected.id,IS:$ibutton.authorized.last.id
	// JSON ex
	// connected_ib:$ibutton.connected
	const ib_dotted = {}
	function onIbuttonChange(update){
		let ib_dotted_update = dotted.dot({ $ibutton: update })
		// nullify all keys on ib_dotted
		clear_object(ib_dotted)
		// apply new update
		Object.assign(ib_dotted, ib_dotted_update)
		Object.assign(syrusjs_variables, ib_dotted);
	}
	// TAIP ex
	// EA_A:$temp.aliases.sensor1.value
	// EA_A:$temp.sensor_list[0].value
	// JSON ex
	// temp:$temp.aliases
	const temp_dotted = {}
	function onTemperatureChange(update) {
		let temp_dotted_update = dotted.dot({ $temp: update })
		// nullify all keys on ib_dotted
		clear_object(temp_dotted)
		// apply update
		Object.assign(temp_dotted, temp_dotted_update);
		Object.assign(syrusjs_variables, temp_dotted);
	}

	function onSerialMessage(message){
		Object.assign(syrusjs_variables, {"$serial.incoming.message": message});
		fireInternalSignal(`@serial.incoming.message`);
	}

	// Establishing all the subcriptions to redis throught the SDK to get the changes on the S4.
	GPS.watchPosition(onGpsData, console.error);
	IOS.watchInputState("*", onInputChange, console.error);
	ECU.watchECUParams(onEcuParamsChange, console.error);
	Geofences.watchGeofences(onGeofenceChange, console.error, { namespace: configuration.namespace });
	Geofences.watchGroups(onGeofenceGroupChange, console.error, { namespace: configuration.namespace });
	Network.onNetworkChange(onNetworkChange, console.error);
	Accelerometer.onMotionChange(onMotionChange, console.error);
	Accelerometer.on(onMotionEvent, console.error);
	System.onSleepOn(onSleepOn, console.error);
	IButton.onIButtonChange(onIbuttonChange, console.error)
	Temperature.onTemperatureChange(onTemperatureChange, console.error)
	Serial.onIncomingMessage(onSerialMessage, console.error)

	// Get the current state of network, ios and ecu, to update the signals and variables
	Network.getNetworks()
		.then(onNetworkChange)
		.catch(console.error);
	Geofences.getAll({ namespace: configuration.namespace })
		.then(onGeofenceChange)
		.catch(console.error);
	IOS.getAll()
		.then(onInputChange)
		.catch(console.error);
	ECU.getECUParams()
		.then(onEcuParamsChange)
		.catch(console.error);

	/**
	 * All the config file are watch in case of any change to restart the app a get the new configuration
	 * TODO: pending improve  get the new config without a restart.
	 */
	var { destinationsPath, configPath, schemaPath } = configuration;
	fs.watch(schemaPath, () => {
		console.log(`detected changes on file ${schemaPath}, restarting app`);
		process.kill(process.pid, "SIGTERM");
	});
	fs.watch(destinationsPath, () => {
		console.log(`detected changes on file ${destinationsPath}, restarting app`);
		process.kill(process.pid, "SIGTERM");
	});
	fs.watch(configPath, () => {
		console.log(`detected changes on file ${configPath}, restarting app`);
		process.kill(process.pid, "SIGTERM");
	});

	// Arbitrary callback to check if syruslang was restarted by powersave mode, this is not called immediatlysince the unix time can be invalid while there is no networkand it is established to be the internal clock of S4
	setTimeout(() => {
		System.getLastWakeUp()
			.then(wakeup => {
				if (wakeup) {
					var newTimestamp = dayjs(wakeup.timestamp);
					var reason = wakeup.wakeup_reason.toLowerCase();
					if (reason == "mot") reason = "motion";
					if (reason == "elapsed_time") reason = "time";
					if (reason == "electrical_rst") reason = "electrical_reset";

					if (newTimestamp.unix() != syrusjs_variables["$power_save.time"]) {
						syrusjs_variables["$power_save.time"] = newTimestamp.unix();
						syrusjs_variables["$power_save.reason"] = wakeup.wakeup_reason;
						fireInternalSignal(`@power_save.wakeup`);
						fireInternalSignal(`@power_save.${reason}`);
					}
				}
			})
			.catch(err => {
				console.error("ERROR getting wakeup reason", err);
			});
	}, 1000 * 15);

	// Interval for sync data from the utils within syrusjs
	// In case redis fails for some reason or the syrus-interface fails for any reason
	// Could be improved this handling this part inside the Syrus4-SDK
	setInterval(() => {
		onNetworkChange();
		IOS.getAll()
			.then(onInputChange)
			.catch(console.error);
		ECU.getECUParams()
			.then(onEcuParamsChange)
			.catch(console.error);
	}, 1000 * 60 * 5);

	// update $modem information periodically
	// TODO: update with publish
	setInterval(async ()=>{
		try {
			var modem = await System.modem()
			for (const key in modem) {
				syrusjs_variables[`$modem.${key.toLowerCase()}`] = modem[key]
			}
		} catch (error) {
			console.error(error)
		}
	}, 1000 * 60 * 1)
};
